// ---------------------------------------------------------------------------
//  M88 - PC-8801 emulator
//	Copyright (C) cisc 1999, 2000.
// ---------------------------------------------------------------------------
//	$Id: memmon.cpp,v 1.10 2000/02/29 12:30:15 cisc Exp $

#include "headers.h"
#include "resource.h"
#include "pc88/pc88.h"
#include "memmon.h"
#include "misc.h"
#include "device_i.h"

using namespace PC8801;

// ---------------------------------------------------------------------------
//	\z/
//
MemoryMonitor::MemoryMonitor()
{
	edlgproc.SetDestination(EDlgProcGate, this);
}

MemoryMonitor::~MemoryMonitor()
{
}

// ---------------------------------------------------------------------------
//	
//
bool MemoryMonitor::Init(PC88* pc88)
{
	if (!MemViewMonitor::Init(MAKEINTRESOURCE(IDD_MEMORY), pc88))
		return false;
	SetLines(0x1000);
	SetUpdateTimer(50);
	return true;
}

inline static void ToHex(char** p, uint d)
{
	static const char hex[] = "0123456789abcdef";

	*(*p)++ = hex[d >> 4];
	*(*p)++ = hex[d & 15];
}

// ---------------------------------------------------------------------------
//	eLXgXV
//
void MemoryMonitor::UpdateText()
{
	char buf[128];
	int a = GetLine() * 0x10;
	
	for (int y=0; y<GetHeight(); y++)
	{
		if (a < 0x10000)
		{
			char* ptrt = buf;
			char* ptrc = buf+48;
			for (int x=0; x<16; x++)
			{
				int d = GetBus()->Read8(a++ & 0xffff);
				ToHex(&ptrt, d & 255);
				*ptrt++ = ' ';
				*ptrc++ = d >= 0x20 && d < 0xfd && d != 0x7f ? d : '.'; 
			}
			Putf("%.4x: %.48s %.16s\n", (a - 0x10) & 0xffff, buf, buf+48);
		}
		else
		{
			Puts("\n");
		}
	}
}

// ---------------------------------------------------------------------------
//	_CAO
//
BOOL MemoryMonitor::DlgProc(HWND hdlg, UINT msg, WPARAM wp, LPARAM lp)
{
	switch (msg)
	{
		POINT p;
	
	case WM_LBUTTONDBLCLK:
		p.x = LOWORD(lp);
		p.y = HIWORD(lp);
		
		if (GetTextPos(&p))
		{
			if (6 <= p.x && p.x <= 5+48)
			{
				editaddr = (GetLine() + p.y) * 16 + (p.x - 6) / 3;
				if (editaddr < 0x10000)
				{
					DialogBox(GetHInst(), MAKEINTRESOURCE(IDD_MEMORY_EDIT), 
								hdlg, DLGPROC((void*) edlgproc));
				}
			}
		}
		return 0;
	}
	return MemViewMonitor::DlgProc(hdlg, msg, wp, lp);
}

// ---------------------------------------------------------------------------
//	e_CAO\
//
BOOL MemoryMonitor::EDlgProc(HWND hdlg, UINT msg, WPARAM wp, LPARAM lp)
{
	char buf[16];
	switch (msg)
	{
	case WM_INITDIALOG:
		wsprintf(buf, "%.4x:", editaddr);
		SetDlgItemText(hdlg, IDC_MEMORY_EDIT_TEXT, buf);
		wsprintf(buf, "%.2x", GetBus()->Read8(editaddr));
		SetDlgItemText(hdlg, IDC_MEMORY_EDITBOX, buf);
		SetFocus(GetDlgItem(hdlg, IDC_MEMORY_EDITBOX));
		SendDlgItemMessage(hdlg, IDC_MEMORY_EDITBOX, EM_SETSEL, 0, -1);
		return 0;

	case WM_COMMAND:
		if (HIWORD(wp) == BN_CLICKED)
		{
			switch (LOWORD(wp))
			{
				char* dum;
			case IDOK:
				GetDlgItemText(hdlg, IDC_MEMORY_EDITBOX, buf, 5);
				if (isxdigit(buf[0]))
					GetBus()->Write8(editaddr, strtoul(buf, &dum, 16));
			case IDCANCEL:
				EndDialog(hdlg, false);
				break;
			}
		}
		return 0;		

	case WM_CLOSE:
		EndDialog(hdlg, false);
		return true;

	default:
		return false;
	}
}

BOOL CALLBACK MemoryMonitor::EDlgProcGate
(MemoryMonitor* about, HWND hwnd, UINT m, WPARAM w, LPARAM l)
{
	return about->EDlgProc(hwnd, m, w, l);
}
